#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
DBUS SERVICE RESTART
"""

from gi.repository import GLib
import time
import subprocess
import logging
import sys
import os
sys.path.insert(1, os.path.join(os.path.dirname(__file__), 'ext', 'velib_python'))

# ************************ToDo

from dbus.mainloop.glib import DBusGMainLoop
from functools import partial
import dbus
from vedbus import VeDbusService
from ve_utils import exit_on_error
from settingsdevice import SettingsDevice
import traceback

# **********************END ToDo

VERSION = '0.6'


INPUTTYPES = [
    'Disabled',
    'DS18B20',
    'BME280',
    'ADS1115',
    'CONTROL',
    'Burglar alarm',
    'Smoke alarm',
    'Fire alarm',
    'CO2 alarm'
]

class SystemBus(dbus.bus.BusConnection):
        def __new__(cls):
                print ("")
                print ("SYSTEMBUS __new__")
                print ("")
                return dbus.bus.BusConnection.__new__(cls, dbus.bus.BusConnection.TYPE_SYSTEM)

class SessionBus(dbus.bus.BusConnection):
        def __new__(cls):
                print ("")
                print ("SESSIONBUS __new__")
                print ("")
                return dbus.bus.BusConnection.__new__(cls, dbus.bus.BusConnection.TYPE_SESSION)

class HandlerMaker(type):
        """ Meta-class for keeping track of all extended classes. """
        def __init__(cls, name, bases, attrs):
                print ("")
                print ("HandlerMaker __init__", "name=", name, "bases=", bases, "attrs=", attrs)
                print ("cls =", cls)
                print ("")
                if not hasattr(cls, 'handlers'):
                        cls.handlers = {}
                else:
                        cls.handlers[cls.type_id] = cls

class PinHandler(object, metaclass=HandlerMaker):
        product_id = 0xFFFF
        _product_name = 'Generic GPIO'
        dbus_name = "CONTROL"

        def __init__(self, bus, base, path, gpio, settings):
                print ("")
                print ("PINHANDLER __init__", "base=", base, "path=", path, "gpio=", gpio, "settings=", settings)
                print ("")

                self.gpio = gpio
                self.path = path
                self.bus = bus
                self.settings = settings

                self._level = 0 # Remember last state

                self.service = VeDbusService("com.victronenergy.control.P-{:d}".format(gpio), bus=bus)

                print ("        --> PINHANDLER SERVICE com.victronenergy.conrtol.P-{:d} at BUS".format(gpio), bus)
                print ("")

                # Add objects required by ve-api
                self.service.add_path('/Mgmt/ProcessName', __file__)
                self.service.add_path('/Mgmt/ProcessVersion', VERSION)
                self.service.add_path('/Mgmt/Connection', path)
                self.service.add_path('/Connection', path)
                self.service.add_path('/DeviceInstance', self.settings['DeviceInstance'])
                self.service.add_path('/ProductId', self.product_id)
                self.service.add_path('/ProductName', self.product_name)
                self.service.add_path('/Connected', 1, writeable=True)
                self.service.add_path('/Status', 0, writeable=True)
                self.service.add_path('/FirmwareVersion',"1.01")

                # Custom name setting
                def _change_name(p, v):
                        # This should fire a change event that will update product_name
                        # below.
                        settings['name'] = v
                        return True

                self.service.add_path('/CustomName', settings['name'], writeable=True,
                        onchangecallback=_change_name)

        @property
        def product_name(self):
                print ("        --> PINHANDLER @property product_name")
                return self.settings['name'] or self._product_name

        @product_name.setter
        def product_name(self, v):
                print ("        --> PINHANDLER @product_name.setter product_name")
                # Some pin types don't have an associated service (Disabled pins for
                # example)
                if self.service is not None:
                        self.service['/ProductName'] = v or self._product_name

        def deactivate(self):
                print ("        --> PINHANDLER deactivate")
                self.service.__del__()
                del self.service
                self.service = None

        @property
        def active(self):
                print ("        --> PINHANDLER active")
                return self.service is not None

        @classmethod
        def createHandler(cls, _type, *args, **kwargs):
                print ("        --> PINHANDLER createHandler", cls,  _type, *args, **kwargs)
                if _type in cls.handlers:
                        return cls.handlers[_type](*args, **kwargs)
                return None

class Control(PinHandler):

        product_id = 0x7353
        _product_name = "CONTROL"
        dbus_name = "control"
        type_id = 3

        def __init__(self, bus, base, path, gpio, settings):
                print("")
                print("CONTROL __init__ bus=", bus, " base=", base, " path=", path, " gpio=", gpio, "settings=", settings)
                print("")
                super(Control, self).__init__(bus, base, path, gpio, settings)
                self.gpio = gpio

#               print ("DEVICE INSTANCE:", settings['DeviceInstance'])


                self.service.add_path('/Gui', 0, writeable=True)
                self.service.add_path('/Control', 0, writeable=True)
                self.service.add_path('/RecBMS', 0, writeable=True)
                self.service.add_path('/Temperature', 0, writeable=True)
                self.service.add_path('/Tank', 0, writeable=True)
                self.service.add_path('/Systemcalc', 0, writeable=True)
                self.service.add_path('/Hostapd', 0, writeable=True)
                self.service.add_path('/Headless', 0, writeable=True)
                self.service.add_path('/Udp', 0, writeable=True)
                self.service.add_path('/Halt', 0, writeable=True)

                self.service['/Mgmt/Connection'] = "CONTROL"
                self.service['/Connection'] = "CONTROL"
                GLib.timeout_add(1000, exit_on_error, self._update)

        @property
        def product_name(self):
                print ("        --> CONTROL @property product_name")
                return self.settings['name'] or self._product_name

        @property
        def rawvalue(self):
                print ("        --> CONTROL @property rawvalue")
                return self.adc_value

        def _update(self):
                # print ("        --> CONTROL _update")

                if (self.service['/Gui'] == 1):
                        self.service['/Gui'] = 0
                        print ('RESTART /Gui')
                        subprocess.Popen("svc -k /service/gui", stdout=subprocess.PIPE, shell=True)

                if (self.service['/Control'] == 1):
                        self.service['/Control'] = 0
                        print ('RESTART /Control')
                        subprocess.Popen("svc -k /service/dbus-control", stdout=subprocess.PIPE, shell=True)

                if (self.service['/RecBMS'] == 1):
                        self.service['/RecBMS'] = 0
                        print ('RESTART /RecBMS')
                        subprocess.Popen("svc -k /service/dbus-recbms", stdout=subprocess.PIPE, shell=True)

                if (self.service['/Temperature'] == 1):
                        self.service['/Temperature'] = 0
                        print ('RESTART /Temperature')
                        subprocess.Popen("svc -k /service/dbus-onewire", stdout=subprocess.PIPE, shell=True)

                if (self.service['/Tank'] == 1):
                        self.service['/Tank'] = 0
                        print ('RESTART /Tank')
                        subprocess.Popen("svc -k /service/dbus-tank", stdout=subprocess.PIPE, shell=True)

                if (self.service['/Udp'] == 1):
                        self.service['/Udp'] = 0
                        print ('RESTART /Udp')
                        subprocess.Popen("svc -k /service/dbus-udp", stdout=subprocess.PIPE, shell=True)

                if (self.service['/Systemcalc'] == 1):
                        self.service['/Systemcalc'] = 0
                        print ('RESTART /Systemcalc')
                        subprocess.Popen("svc -k /service/dbus-systemcalc-py", stdout=subprocess.PIPE, shell=True)

                if (self.service['/Hostapd'] == 1):
                        self.service['/Hostapd'] = 0
                        print ('RESTART /Hostapd')
                        subprocess.Popen("svc -k /service/hostapd", stdout=subprocess.PIPE, shell=True)

                if (self.service['/Headless'] == 1):
                        self.service['/Headless'] = 0
                        print ('RESTART /Headless')
                        subprocess.Popen("touch /etc/venus/headless", stdout=subprocess.PIPE, shell=True)

                if (self.service['/Halt'] == 1):
                        self.service['/Halt'] = 0
                        print ('HALT system')
                        subprocess.Popen("shutdown -r now", stdout=subprocess.PIPE, shell=True)

                return True


# print (os.path.dirname(__file__))
from vedbus import VeDbusService

def dbusconnection():
        print ("DBUSCONNECTION")
        return SessionBus() if 'DBUS_SESSION_BUS_ADDRESS' in os.environ else SystemBus()


# === All code below is to simply run it from the commandline for debugging purposes ===

def main():
        print (" ********************************************* ")
        print ("        M A I N    S T A R T E D ")
        print (" ********************************************* ")
        print ("")

#       logging.basicConfig(level=logging.DEBUG)

    # Have a mainloop, so we can send/receive asynchronous calls to and from dbus
        DBusGMainLoop(set_as_default=True)

        services = {}
        def register_gpio(path, gpio, bus, settings):
                print ("")
                print ("MAIN    ->      register_gpio, path=", path, " bus=", bus, " settings=", settings)
                print ("")
                _type = settings['inputtype']
                _i = settings['DeviceInstance']
                gpio = _i

                handler = PinHandler.createHandler(_type, bus, "/opt/victronenergy/dbus-control/", path, gpio, settings)
                services[gpio] = handler
                # print (services)
        def handle_setting_change(inp, setting, old, new):
                print ("        --> MAIN handle_setting_change inp=", inp, " setting=", setting, " old=", old, " new=", new)

                if setting == 'inputtype':
                        if new:
                                # Get current bus and settings objects, to be reused
                                service = services[inp]
                                bus, settings = service.bus, service.settings
                                # Input enabled. If already enabled, unregister the old one first.
                                if pulses.registered(inp):
                                        unregister_gpio(inp)
                                register_gpio(inputs[inp], inp, bus, settings)
                        elif old:
                                # Input disabled
                                unregister_gpio(inp)
                elif setting == 'name':
                        services[inp]._product_name = new

#               elif setting == 'gpio':
#                    # Get current bus and settings objects, to be reused
#                    service = services[inp]
#                    bus, settings = service.bus, service.settings
#                   settings['DeviceInstance'] = new
#                    # Input enabled. If already enabled, unregister the old one first.
#                   services[inp].gpio = new
#                    register_gpio(inputs[new], new, bus, settings)
#                    unregister_gpio(inp)

        i = 1
        inp = 1
        pth = 0x48
        supported_settings = {
                        'inputtype': ['/Settings/Control/{}/Type'.format(inp), 3, 0, len(INPUTTYPES)],
                        'name': ['/Settings/Control/{}/CustomName'.format(inp), '', '', ''],
                        'DeviceInstance': ['/Settings/Control/{}/DeviceInstance'.format(inp), i, 0, 99],
                }
        bus = dbusconnection()
        sd = SettingsDevice(bus, supported_settings, partial(handle_setting_change, inp), timeout=10)
        register_gpio(pth, i, bus, sd)

        print ("MAIN LOOP")
        print ("")
        mainloop = GLib.MainLoop()
        mainloop.run()


if __name__ == "__main__":
    main()